
#ifndef __CONTEXTFREEGRAMMAR_H
#define __CONTEXTFREEGRAMMAR_H


#include <iostream>
#include <vector>
#include <map>
#include <set>
#include <string>
#include <algorithm>
#include <queue>

namespace ContextFreeGrammars
{

    class Symbol
    {
        public:
            std::string Name;
            Symbol(std::string Name);
            virtual ~Symbol();
    };

    class Terminal : public Symbol
    {
        public:
            Terminal(std::string Name);
            virtual ~Terminal();
    };

    class Production
    {
        public:
            Production();
            virtual ~Production();
            std::vector<Symbol*> produced;
    };

    class NonTerminal : public Symbol
    {
        public:
            std::vector<Production*> productions;
            NonTerminal(std::string Name);
            virtual ~NonTerminal();
            void AddOrDeleteProduction(Production* production);
    };

    class ContextFreeGrammar
    {
        public:
            std::map<std::string, Symbol*> symbols;

            std::vector<NonTerminal*> nonterminals;
            std::vector<Terminal*> terminals;
            NonTerminal* startsymbol;

            ContextFreeGrammar();
            ContextFreeGrammar(const ContextFreeGrammar& cfg);
            ContextFreeGrammar& operator=(const ContextFreeGrammar& cfg);
            virtual ~ContextFreeGrammar();
            void Clear();
            Terminal* AddTerminal(std::string Name);
            NonTerminal* AddNonTerminal(std::string Name);
            void RemoveSymbol(Symbol* symbol);


            bool LoadFromStream(std::istream& is);
            void WriteToStream(std::ostream& os);

            std::vector<NonTerminal*> FindUnproductiveNonTerminals() const;
            void RemoveUnproductiveNonTerminals();
            std::vector<Symbol*> FindUnreachableSymbols() const;
            void RemoveUnreachableSymbols();

            std::set<const NonTerminal*> FindEpsilonProducers() const;
        private:
            void AddEpsilonCleanedProductions(Production* prod, int position, std::vector<Symbol*>& stack,
                 std::set<const NonTerminal*>& epsProducers, NonTerminal* nt, std::map<Symbol*, Symbol*>& correspondingsymbol);
        public:
            ContextFreeGrammar EpsilonElimination() const;
            ContextFreeGrammar UnitProductionElimination() const;
            ContextFreeGrammar ChomskyNormalForm() const;

            void Produce(int depth, const std::vector<const Symbol*>& stack, void callback(const std::vector<const Symbol*>&)) const;
            void Produce(int depth, void callback(const std::vector<const Symbol*>&)) const;

            bool Producable(const std::vector<std::string>& w) const;
            bool Producable(const std::vector<Symbol*>& w) const;
    };

}

#endif
